"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "AbstractGraphQLFilesystem", {
  enumerable: true,
  get: function () {
    return _index.AbstractGraphQLFilesystem;
  }
});
Object.defineProperty(exports, "EnumFormat", {
  enumerable: true,
  get: function () {
    return _types.EnumFormat;
  }
});
Object.defineProperty(exports, "ExportFormat", {
  enumerable: true,
  get: function () {
    return _types.ExportFormat;
  }
});
exports.Builder = void 0;

var _events = require("events");

var _path = require("path");

var _graphql = require("graphql");

var _chalk = _interopRequireDefault(require("chalk"));

var _fsExtra = require("fs-extra");

var _graphqlConfig = require("graphql-config");

var _graphqlConfigUtilities = require("graphql-config-utilities");

var _graphqlToolUtilities = require("graphql-tool-utilities");

var _index = require("./filesystem/index.js");

var _index2 = require("./print/index.js");

var _types = require("./types.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function () { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

class Builder extends _events.EventEmitter {
  // workspace graphql configuration
  // see: https://github.com/prisma/graphql-config
  // projectName -> {filePath -> document}
  // NOTE: projectName can be undefined for nameless graphql-config projects
  constructor({
    cwd,
    ...options
  }) {
    super();
    this.options = void 0;
    this.config = void 0;
    this.documentMapByProject = new Map();
    this.filesystem = void 0;

    this.handleDocumentUpdate = async (filePath, projectConfig) => {
      try {
        await this.updateDocumentForFile(filePath, projectConfig);
      } catch (error) {
        this.emit('error', error);
        return;
      }

      await this.generateDocumentTypes();
    };

    this.handleDocumentDelete = async (filePath, projectConfig) => {
      const documents = this.documentMapByProject.get(projectConfig.projectName);

      if (documents) {
        documents.delete(filePath);
      }

      await this.generateDocumentTypes();
    };

    this.handleSchemaUpdate = async schemaPath => {
      try {
        this.emit('start:schema');
        await this.generateSchemaTypes(schemaPath);
        this.emit('end:schema');
        await this.generateDocumentTypes();
      } catch (error) {// intentional noop
      }
    };

    this.options = options;
    this.config = options.config ? options.config : (0, _graphqlConfig.getGraphQLConfig)(cwd ? (0, _path.resolve)(cwd) : undefined);
  }

  once(event, handler) {
    return super.once(event, handler);
  }

  on(event, handler) {
    return super.on(event, handler);
  }

  emit(event, ...args) {
    return super.emit(event, ...args);
  }

  async run({
    watch: watchGlobs = false,
    graphQLFilesystem
  } = {}) {
    let schemaPaths;

    try {
      schemaPaths = (0, _graphqlConfigUtilities.getGraphQLSchemaPaths)(this.config);
    } catch (error) {
      this.emit('error', error);
      return;
    }

    const filesystem = graphQLFilesystem !== null && graphQLFilesystem !== void 0 ? graphQLFilesystem : await defaultFilesytem();
    this.filesystem = filesystem;

    if (watchGlobs) {
      filesystem.on('change:document', this.handleDocumentUpdate).on('delete:document', this.handleDocumentDelete).on('change:schema', this.handleSchemaUpdate); // wait for all watchers to be ready

      await filesystem.watch(this.config);
    }

    try {
      this.emit('start:schema');
      await Promise.all(schemaPaths.map(schemaPath => this.generateSchemaTypes(schemaPath)));
      this.emit('end:schema');
    } catch (error) {
      this.emit('error', error);
      return;
    }

    try {
      await Promise.all((0, _graphqlConfigUtilities.getGraphQLProjects)(this.config).map(projectConfig => this.updateDocumentsForProject(filesystem, projectConfig)));
    } catch (error) {
      this.emit('error', error);
      return;
    }

    await this.generateDocumentTypes();
  }

  stop() {
    var _this$filesystem;

    (_this$filesystem = this.filesystem) === null || _this$filesystem === void 0 ? void 0 : _this$filesystem.dispose();
  }

  async generateSchemaTypes(schemaPath) {
    const projectConfig = (0, _graphqlConfigUtilities.getGraphQLProjectForSchemaPath)(this.config, schemaPath);
    const schemaTypesPath = getSchemaTypesPath(projectConfig, this.options);
    const definitions = (0, _index2.generateSchemaTypes)(projectConfig.getSchema(), this.options);
    await (0, _fsExtra.mkdirp)(schemaTypesPath);
    await Promise.all(Array.from(definitions.entries()).map(([fileName, definition]) => (0, _fsExtra.writeFile)((0, _path.join)(schemaTypesPath, fileName), definition)));
    this.emit('build:schema', {
      schemaPath,
      schemaTypesPath
    });
  }

  async generateDocumentTypes() {
    this.emit('start:docs');
    this.checkForDuplicateOperations();
    this.checkForDuplicateFragments();
    await Promise.all(Array.from(this.documentMapByProject.entries()).map(([projectName, documents]) => this.generateDocumentTypesForProject(this.config.getProjectConfig(projectName), documents)));
    this.emit('end:docs');
  }

  checkForDuplicateOperations() {
    getDuplicateOperations(this.documentMapByProject).forEach(({
      projectName,
      duplicates
    }) => {
      if (duplicates.length) {
        duplicates.forEach(({
          operationName,
          filePaths
        }) => {
          const message = `GraphQL operations must have a unique name. The operation ${_chalk.default.bold(operationName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${_chalk.default.bold(projectName)})` : ''}`;
          this.emit('error', new Error(message));
        });
      }
    });
  }

  checkForDuplicateFragments() {
    getDuplicateFragments(this.documentMapByProject).forEach(({
      projectName,
      duplicates
    }) => {
      if (duplicates.length) {
        duplicates.forEach(({
          fragmentName,
          filePaths
        }) => {
          const message = `GraphQL fragments must have a unique name. The fragment ${_chalk.default.bold(fragmentName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${_chalk.default.bold(projectName)})` : ''}`;
          this.emit('error', new Error(message));
        });
      }
    });
  }

  async generateDocumentTypesForProject(projectConfig, documents) {
    let ast;

    try {
      ast = (0, _graphqlToolUtilities.compile)(projectConfig.getSchema(), (0, _graphql.concatAST)(Array.from(documents.values())));
    } catch (error) {
      this.emit('error', error);
      return;
    }

    try {
      await Promise.all(Array.from(groupOperationsAndFragmentsByFile(ast).values()).map(file => this.writeDocumentFile(file, ast, projectConfig)));
    } catch (error) {// intentional noop
    }
  }

  async writeDocumentFile(file, ast, projectConfig) {
    const definitionPath = `${file.path}.d.ts`;
    await (0, _fsExtra.writeFile)(definitionPath, this.getDocumentDefinition(file, ast, projectConfig));
    this.emit('build:docs', {
      documentPath: file.path,
      definitionPath,
      operation: file.operation,
      fragments: file.fragments
    });
  }

  getDocumentDefinition(file, ast, projectConfig) {
    try {
      return (0, _index2.printDocument)(file, ast, {
        enumFormat: this.options.enumFormat,
        exportFormat: this.options.exportFormat,
        addTypename: this.options.addTypename,
        schemaTypesPath: getSchemaTypesPath(projectConfig, this.options)
      });
    } catch ({
      message
    }) {
      const error = new Error(`Error in ${file.path}: ${message[0].toLowerCase()}${message.slice(1)}`);
      this.emit('error', error);
      throw error;
    }
  }

  async updateDocumentsForProject(filesystem, projectConfig) {
    const filePaths = await filesystem.getGraphQLProjectIncludedFilePaths(projectConfig);
    return Promise.all(filePaths.map(filePath => this.updateDocumentForFile(filePath, projectConfig)));
  }

  async updateDocumentForFile(filePath, projectConfig) {
    const contents = await (0, _fsExtra.readFile)(filePath, 'utf8');
    return this.setDocumentForFilePath(filePath, projectConfig, contents);
  }

  setDocumentForFilePath(filePath, projectConfig, contents) {
    let documents = this.documentMapByProject.get(projectConfig.projectName);

    if (!documents) {
      documents = new Map();
      this.documentMapByProject.set(projectConfig.projectName, documents);
    }

    if (contents.trim().length === 0) {
      return undefined;
    }

    const document = (0, _graphql.parse)(new _graphql.Source(contents, filePath));
    documents.set(filePath, document);
    return document;
  }

}

exports.Builder = Builder;

function getSchemaTypesPath(projectConfig, options) {
  if (typeof projectConfig.extensions.schemaTypesPath === 'string') {
    return (0, _graphqlConfigUtilities.resolvePathRelativeToConfig)(projectConfig, projectConfig.extensions.schemaTypesPath);
  }

  return (0, _graphqlConfigUtilities.resolvePathRelativeToConfig)(projectConfig, (0, _path.join)(options.schemaTypesPath, `${projectConfig.projectName ? `${projectConfig.projectName}-` : ''}types`));
}

async function defaultFilesytem() {
  const {
    DefaultGraphQLFilesystem
  } = await Promise.resolve().then(() => _interopRequireWildcard(require('./filesystem/default-graphql-filesystem')));
  return new DefaultGraphQLFilesystem();
}

function groupOperationsAndFragmentsByFile({
  operations,
  fragments
}) {
  return Object.values(operations).concat(Object.values(fragments)).reduce((map, item) => {
    if (!item.filePath) {
      return map;
    }

    let file = map.get(item.filePath);

    if (!file) {
      file = {
        path: item.filePath,
        operation: undefined,
        fragments: []
      };
      map.set(item.filePath, file);
    }

    if ((0, _graphqlToolUtilities.isOperation)(item)) {
      file.operation = item;
    } else {
      file.fragments.push(item);
    }

    return map;
  }, new Map());
}

function getDuplicateOperations(documentsMapByProject) {
  return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
    return {
      projectName,
      duplicates: getDuplicateProjectOperations(documents)
    };
  });
}

function getDuplicateFragments(documentsMapByProject) {
  return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
    return {
      projectName,
      duplicates: getDuplicateProjectFragments(documents)
    };
  });
}

function getDuplicateProjectOperations(documents) {
  const operations = new Map();
  Array.from(documents.entries()).forEach(([filePath, document]) => {
    document.definitions.filter(isOperationDefinition).forEach(definition => {
      const {
        name
      } = definition;

      if (name && name.value) {
        const map = operations.get(name.value);

        if (map) {
          map.add(filePath);
        } else {
          operations.set(name.value, new Set([filePath]));
        }
      }
    });
  });
  return Array.from(operations.entries()).filter(([, filePaths]) => filePaths.size > 1).map(([operationName, filePath]) => {
    return {
      operationName,
      filePaths: Array.from(filePath)
    };
  });
}

function getDuplicateProjectFragments(documents) {
  const fragments = new Map();
  Array.from(documents.entries()).forEach(([filePath, document]) => {
    document.definitions.filter(isFragmentDefinition).forEach(definition => {
      const {
        name
      } = definition;

      if (name && name.value) {
        const map = fragments.get(name.value);

        if (map) {
          map.add(filePath);
        } else {
          fragments.set(name.value, new Set([filePath]));
        }
      }
    });
  });
  return Array.from(fragments.entries()).filter(([, filePaths]) => filePaths.size > 1).map(([fragmentName, filePath]) => {
    return {
      fragmentName,
      filePaths: Array.from(filePath)
    };
  });
}

function isOperationDefinition(definition) {
  return definition.kind === 'OperationDefinition';
}

function isFragmentDefinition(definition) {
  return definition.kind === 'FragmentDefinition';
}