"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.generateSchemaTypes = generateSchemaTypes;

var t = _interopRequireWildcard(require("@babel/types"));

var _changeCase = require("change-case");

var _graphql = require("graphql");

var _utilities = require("../utilities.js");

var _types2 = require("../../types.js");

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function () { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const generate = require('@babel/generator').default;

function generateSchemaTypes(schema, options = {}) {
  const importFileBody = [];
  const exportFileBody = [];
  const definitions = new Map();

  for (const type of Object.values(schema.getTypeMap())) {
    if (!(0, _graphql.isInputType)(type) || type.name.startsWith('__')) {
      continue;
    }

    if ((0, _graphql.isScalarType)(type) && Object.prototype.hasOwnProperty.call(_utilities.scalarTypeMap, type.name)) {
      continue;
    }

    if ((0, _graphql.isEnumType)(type)) {
      const enumType = tsEnumForType(type, options);
      definitions.set(`${enumType.id.name}.ts`, generate(t.file(t.program([t.exportNamedDeclaration(enumType, [])]), [], [])).code);
      importFileBody.unshift(t.importDeclaration([t.importSpecifier(enumType.id, enumType.id)], t.stringLiteral(`./${enumType.id.name}`)));
      exportFileBody.unshift(t.exportNamedDeclaration(null, [t.exportSpecifier(enumType.id, enumType.id)]));
    } else if ((0, _graphql.isScalarType)(type)) {
      const {
        customScalars = {}
      } = options;
      const customScalarDefinition = customScalars[type.name];
      const scalarType = tsScalarForType(type, customScalarDefinition);

      if (customScalarDefinition && customScalarDefinition.package) {
        importFileBody.unshift(t.importDeclaration([t.importSpecifier(t.identifier(makeCustomScalarImportNameSafe(customScalarDefinition.name, type.name)), t.identifier(customScalarDefinition.name))], t.stringLiteral(customScalarDefinition.package)));
        exportFileBody.unshift(t.exportNamedDeclaration(scalarType, []));
      } else {
        exportFileBody.push(t.exportNamedDeclaration(scalarType, []));
      }
    } else {
      exportFileBody.push(t.exportNamedDeclaration(tsInputObjectForType(type), []));
    }
  } // A blank file is ambiguous - its not clear if it is a script or a module.
  // If the file would be blank then give it an empty `export {}` to make in
  // unambiguously an es module file.
  // This ensures the generated files passes TypeScript type checking in
  // "isolatedModules" mode.


  if (exportFileBody.length === 0) {
    exportFileBody.push(t.exportNamedDeclaration(null, []));
  }

  return definitions.set('index.ts', generate(t.file(t.program(importFileBody.concat(exportFileBody)), [], [])).code);
}

function tsTypeForInputType(type) {
  const unwrappedType = (0, _graphql.isNonNullType)(type) ? type.ofType : type;
  let tsType;

  if ((0, _graphql.isListType)(unwrappedType)) {
    const tsTypeOfContainedType = tsTypeForInputType(unwrappedType.ofType);
    tsType = t.tsArrayType(t.isTSUnionType(tsTypeOfContainedType) ? t.tsParenthesizedType(tsTypeOfContainedType) : tsTypeOfContainedType);
  } else if ((0, _graphql.isScalarType)(unwrappedType)) {
    tsType = _utilities.scalarTypeMap[unwrappedType.name] || t.tsTypeReference(t.identifier(unwrappedType.name));
  } else {
    tsType = t.tsTypeReference(t.identifier(unwrappedType.name));
  }

  return (0, _graphql.isNonNullType)(type) ? tsType : t.tsUnionType([tsType, t.tsNullKeyword()]);
}

function tsInputObjectForType(type) {
  const fields = Object.entries(type.getFields()).map(([name, field]) => {
    const property = t.tsPropertySignature(t.identifier(name), t.tsTypeAnnotation(tsTypeForInputType(field.type)));
    property.optional = !(0, _graphql.isNonNullType)(field.type);
    return property;
  });
  return t.tsInterfaceDeclaration(t.identifier(type.name), null, null, t.tsInterfaceBody(fields));
}

function tsScalarForType(type, customScalarDefinition) {
  let alias;

  if (customScalarDefinition && customScalarDefinition.package) {
    alias = t.tsTypeReference(t.identifier(makeCustomScalarImportNameSafe(customScalarDefinition.name, type.name)));
  } else if (customScalarDefinition) {
    alias = t.tsTypeReference(t.identifier(customScalarDefinition.name));
  } else {
    alias = t.tsStringKeyword();
  }

  return t.tsTypeAliasDeclaration(t.identifier(type.name), null, alias);
}

function makeCustomScalarImportNameSafe(importName, typeName) {
  return `__${typeName}__${importName}`;
}

function tsEnumForType(type, {
  enumFormat
}) {
  return t.tsEnumDeclaration(t.identifier(type.name), type.getValues().map(value => t.tsEnumMember(t.identifier(enumMemberName(value.name, enumFormat)), t.stringLiteral(value.name))));
}

function enumMemberName(name, format) {
  switch (format) {
    case _types2.EnumFormat.CamelCase:
      return (0, _changeCase.camelCase)(name);

    case _types2.EnumFormat.PascalCase:
      return (0, _changeCase.pascalCase)(name);

    case _types2.EnumFormat.SnakeCase:
      return (0, _changeCase.snakeCase)(name);

    case _types2.EnumFormat.ScreamingSnakeCase:
      return (0, _changeCase.snakeCase)(name).toUpperCase();

    default:
      return name;
  }
}